"""
Abstractions for nodes representation.
"""

from __future__ import annotations

from abc import ABC, abstractmethod
from copy import deepcopy
from itertools import chain, count, cycle
from typing import Callable, Dict, Iterable, Iterator, List, Sequence, Set, Union, cast

import netunicorn.base

from .architecture import Architecture
from .environment_definitions import _available_environment_definitions
from .types import NodeProperty, NodeRepresentation, NodesRepresentation


class Node:
    """
    Represents a single node in a pool of nodes.

    :param name: name of the node
    :param properties: custom properties of the node
    :param architecture: node architecture
    """

    def __init__(
        self,
        name: str,
        properties: Dict[str, NodeProperty],
        architecture: Architecture = Architecture.UNKNOWN,
    ):
        self.name: str = name
        """
        Node name.
        """

        self.properties: Dict[str, NodeProperty] = properties
        """
        Node properties. Could be used to store custom information about the node.
        """

        self.additional_properties: Dict[str, NodeProperty] = {}
        """
        Additional node properties, often used for internal purposes and not to be exposed to the user.
        """

        self.architecture: Architecture = architecture
        """
        Node architecture.
        """

        self.available_environments: Set[type] = self._infer_environments()
        """
        Supported environments for the node (see :py:mod:`netunicorn.base.environment_definitions`).
        """

    def _infer_environments(self) -> Set[type]:
        result = set()
        # noinspection PyBroadException
        try:
            environments = self.properties.get(
                "netunicorn-environments", _available_environment_definitions.keys()
            )
            if hasattr(environments, "__iter__"):
                for environment_name in cast(Iterable[str], environments):
                    if environment_name in _available_environment_definitions:
                        result.add(_available_environment_definitions[environment_name])
        except Exception:
            return set(_available_environment_definitions.values())
        return result

    def __getitem__(self, item: str) -> NodeProperty:
        """
        Returns a node property by name.

        :param item: name of the property
        :return: property value
        """
        return self.properties.get(item, None)

    def __iter__(self) -> Iterator[NodeProperty]:
        raise TypeError("Node is not iterable")

    def __setitem__(self, key: str, value: NodeProperty) -> None:
        """
        Sets a node property.

        :param key: name of the property
        :param value: property value
        """
        self.properties[key] = value

    def __str__(self) -> str:
        return self.name

    def __repr__(self) -> str:
        return self.name

    def __eq__(self, other: object) -> bool:
        """
        Checks if two nodes are equal.

        :param other: other node
        :return: True if the nodes have all parameters equal, False otherwise
        """
        return (
            isinstance(other, Node)
            and self.name == other.name
            and self.properties == other.properties
            and self.additional_properties == other.additional_properties
            and self.architecture == other.architecture
        )

    def __json__(self) -> NodeRepresentation:
        return {
            "name": self.name,
            "properties": self.properties,
            "additional_properties": self.additional_properties,
            "architecture": self.architecture.value,
        }

    @classmethod
    def from_json(cls, data: NodeRepresentation) -> Node:
        """
        Returns an instance of the object from a JSON representation.

        :param data: JSON representation of the object
        :return: instance of the object
        """
        instance = cls(
            data["name"], data["properties"], Architecture(data["architecture"])
        )
        instance.additional_properties = data["additional_properties"]
        return instance


class Nodes(ABC):
    """
    A base class that represents a pool of nodes. Not to be used directly, but to be inherited from.
    """

    @abstractmethod
    def __json__(self) -> NodesRepresentation:
        """
        Returns a JSON representation of the object.
        """
        pass

    @staticmethod
    def dispatch_and_deserialize(data: NodesRepresentation) -> Nodes:
        """
        Deserializes a JSON representation of the object and returns an instance of the object.

        :param data: JSON representation of the object
        :return: instance of the object
        """

        cls: Nodes = getattr(netunicorn.base.nodes, data["node_pool_type"])
        return cls.from_json(data["node_pool_data"])

    @classmethod
    @abstractmethod
    def from_json(
        cls, data: List[Union[NodeRepresentation, NodesRepresentation]]
    ) -> Nodes:
        """
        Class-specific implementation of deserialization from JSON.

        :param data: JSON representation of the object
        :return: instance of the object
        """
        pass

    @abstractmethod
    def __str__(self) -> str:
        pass

    @abstractmethod
    def __len__(self) -> int:
        pass

    @abstractmethod
    def __getitem__(self, item: int) -> Union[Node, Nodes]:
        pass

    @abstractmethod
    def filter(self, function: Callable[[Node], bool]) -> Nodes:
        """
        Returns a pool of nodes that match the given filter function.

        :param function: filter function returning True if the node should be included in the result
        :return: pool of nodes
        """
        pass

    @abstractmethod
    def take(self, count: int) -> Sequence[Node]:
        """
        Returns a sequence of nodes consisting of the first `count` nodes.

        :param count: number of nodes to take
        :return: sequence of nodes
        """
        pass

    @abstractmethod
    def skip(self, count: int) -> Nodes:
        """
        Returns a pool of nodes consisting of the nodes after the first 'count' nodes.

        :param count: number of nodes to skip
        :return: pool of nodes
        """
        pass

    @abstractmethod
    def set_property(self, name: str, value: NodeProperty) -> Nodes:
        """
        Sets a property for all nodes in the pool.

        :param name: name of the property
        :param value: property value
        :return: self
        """
        pass


class CountableNodePool(Nodes):
    """
    Represents a pool of nodes that contains a fixed number of nodes.

    :param nodes: list of nodes
    """

    def __init__(self, nodes: List[Union[Node, Nodes]]):
        self.nodes = nodes
        """
        Nodes in the pool.
        """

    def __json__(self) -> NodesRepresentation:
        return {
            "node_pool_type": self.__class__.__name__,
            "node_pool_data": [x.__json__() for x in self.nodes],
        }

    @classmethod
    def from_json(
        cls, data: List[Union[NodeRepresentation, NodesRepresentation]]
    ) -> CountableNodePool:
        """
        Returns an instance of the object from a JSON representation.

        :param data: JSON representation of the object
        :return: instance of the object
        """
        nodes: List[Union[Node, Nodes]] = []
        for element in data:
            if "node_pool_type" in element:
                # we know this is a NodesRepresentation
                nodes_representation_element = cast(NodesRepresentation, element)
                nodes.append(
                    Nodes.dispatch_and_deserialize(nodes_representation_element)
                )
            else:
                nodes.append(Node.from_json(element))
        return cls(nodes)

    def __str__(self) -> str:
        return str(self.nodes)

    def __len__(self) -> int:
        return len(self.nodes)

    def __getitem__(self, key: int) -> Union[Node, Nodes]:
        return self.nodes[key]

    def __setitem__(
        self, key: int, value: Union[Node, CountableNodePool, UncountableNodePool]
    ) -> None:
        self.nodes[key] = value

    def pop(self, index: int) -> Union[Node, Nodes]:
        """
        Removes and returns the node at the given index.

        :param index: index of the node to remove
        :return: popped node
        """
        return self.nodes.pop(index)

    def __repr__(self) -> str:
        return str(self.nodes)

    def filter(self, function: Callable[[Node], bool]) -> CountableNodePool:
        nodes: List[Union[Node, Nodes]] = []
        for node in self.nodes:
            if isinstance(node, Node):
                if function(node):
                    nodes.append(node)
            else:
                filtered_nodes = node.filter(function)
                if len(filtered_nodes) > 0:
                    nodes.append(filtered_nodes)
        return CountableNodePool(nodes)

    def take(self, count: int) -> Sequence[Node]:
        iterator: Iterator[Node] = chain.from_iterable(
            cast(
                Iterable[Iterable[Node]],
                ([x] if isinstance(x, Node) else x for x in self.nodes),
            )
        )
        nodes = []
        for _ in range(count):
            try:
                nodes.append(next(iterator))
            except StopIteration:
                break
        return nodes

    def skip(self, count: int) -> CountableNodePool:
        if count > len(self.nodes):
            print(
                f"Warning: asked to skip {count} nodes, but pool length is {len(self.nodes)}. "
                f"Returning empty pool."
            )
            return CountableNodePool([])
        return CountableNodePool(self.nodes[count:])

    def set_property(self, name: str, value: NodeProperty) -> CountableNodePool:
        for node in self.nodes:
            if isinstance(node, Node):
                node.properties[name] = value
            else:
                node.set_property(name, value)
        return self


class UncountableNodePool(Nodes):
    """
    | Represents a pool of nodes that is not countable (i.e., contains possibly infinite amount of nodes, e.g. cloud provider).
    | In the current implementation cannot have Nodes as elements.

    :param node_template: list of nodes that will be used as a template for generating new nodes
    """

    def __init__(self, node_template: List[Node]):
        self._node_template = node_template
        """
        Node template used for generating new nodes.
        """

        self._nodes = cycle(node_template)
        """
        An iterator over the node template.
        """

        self._counter = count(start=1, step=1)
        """
        Node name counter.
        """

    def __str__(self) -> str:
        return str(
            f"<Uncountable node pool with next node template: {self._node_template}>"
        )

    def __repr__(self) -> str:
        return str(self)

    def __iter__(self) -> Iterator[Node]:
        return self

    def __next__(self) -> Node:
        """
        Generate and returns the next node in the pool.

        :return: next node
        """
        node = deepcopy(next(self._nodes))
        node.name += str(next(self._counter))
        return node

    def __getitem__(self, key: int) -> Node:
        return self._node_template[key]

    def __setitem__(self, key: int, value: Node) -> None:
        self._node_template[key] = value

    def __len__(self) -> int:
        return len(self._node_template)

    def filter(self, function: Callable[[Node], bool]) -> UncountableNodePool:
        return UncountableNodePool([x for x in self._node_template if function(x)])

    def take(self, count: int) -> Sequence[Node]:
        nodes = []
        for _ in range(count):
            nodes.append(next(self))
        return nodes

    def skip(self, count: int) -> UncountableNodePool:
        count = count % len(self._node_template)
        for _ in range(count):
            next(self._nodes)
        return self

    def __json__(self) -> NodesRepresentation:
        return {
            "node_pool_type": self.__class__.__name__,
            "node_pool_data": [x.__json__() for x in self._node_template],
        }

    @classmethod
    def from_json(
        cls, data: List[Union[NodeRepresentation, NodesRepresentation]]
    ) -> UncountableNodePool:
        for x in data:
            if "node_pool_type" in x:
                raise ValueError("UncountableNodePool cannot have Nodes as elements.")
        # now we have only NodeRepresentation in the list
        node_representation_data = cast(List[NodeRepresentation], data)
        return cls([Node.from_json(x) for x in node_representation_data])

    def set_property(self, name: str, value: NodeProperty) -> UncountableNodePool:
        for node in self._node_template:
            node.properties[name] = value
        return self



"""
Single deployment of pipeline on node.
"""
from __future__ import annotations

from base64 import b64decode
from copy import deepcopy
from typing import List, Optional, cast

import netunicorn.base.environment_definitions

from .nodes import Node
from .pipeline import Pipeline
from .task import Task, TaskDispatcher
from .types import DeploymentRepresentation
from .utils import SerializedPipelineType

try:
    import cloudpickle  # it's needed only on client side, but this module is also imported on engine side
    import netunicorn.library

    cloudpickle.register_pickle_by_value(netunicorn.library)
except ImportError:
    pass


class Deployment:
    """
    Single deployment of pipeline on node.

    :param node: Node to deploy pipeline on
    :param pipeline: Pipeline to deploy
    :param keep_alive_timeout_minutes: time to wait for executor update before timeout
    :param cleanup: whether to remove artifacts (e.g., Docker image and containers) after execution
    """

    def __init__(
        self,
        node: Node,
        pipeline: Pipeline,
        keep_alive_timeout_minutes: int = 10,
        cleanup: bool = True,
    ):
        self.node: Node = node
        """
        Node to deploy pipeline on
        """

        self.prepared: bool = False
        """
        if False, deployment is not prepared yet or failed during preparation
        """

        self.executor_id: str = ""
        """
        ID of executor on node
        """

        self.error: Optional[Exception] = None
        """
        if deployment failed, this field contains error
        """

        self.pipeline: SerializedPipelineType = b""
        """
        Serialized Pipeline to be deployed
        """

        self.environment_definition: netunicorn.base.environment_definitions.EnvironmentDefinition = deepcopy(
            pipeline.environment_definition
        )
        """
        Environment definition to use for deployment
        """

        self.keep_alive_timeout_minutes: int = keep_alive_timeout_minutes
        """
        time to wait for executor update before timeout
        """

        self.cleanup: bool = cleanup
        """
        if True, corresponding artifacts (Docker image and containers) will be removed after execution
        """

        self._validate_deployment(node, pipeline)

        pipeline = deepcopy(pipeline)

        for i, element in enumerate(pipeline.tasks):
            pipeline.tasks[i] = [
                x.dispatch(node) if isinstance(x, TaskDispatcher) else x
                for x in element
            ]
            tasks_list: List[Task] = cast(List[Task], pipeline.tasks[i])
            for x in tasks_list:
                self.environment_definition.commands.extend(x.requirements)

        self.pipeline = cloudpickle.dumps(pipeline)

    @staticmethod
    def _validate_deployment(node: Node, pipeline: Pipeline) -> None:
        if type(pipeline.environment_definition) not in node.available_environments:
            raise ValueError(
                f"Node {node.name} does not support environment {type(pipeline.environment_definition).__name__}"
            )

    def __str__(self) -> str:
        return f"Deployment: Node={self.node.name}, executor_id={self.executor_id}, prepared={self.prepared}"

    def __repr__(self) -> str:
        return self.__str__()

    def __json__(self) -> DeploymentRepresentation:
        return {
            "node": self.node.__json__(),
            "prepared": self.prepared,
            "executor_id": self.executor_id,
            "error": str(self.error) if self.error else None,
            "pipeline": self.pipeline,
            "keep_alive_timeout_minutes": self.keep_alive_timeout_minutes,
            "cleanup": self.cleanup,
            "environment_definition": self.environment_definition.__json__(),
            "environment_definition_type": self.environment_definition.__class__.__name__,
        }

    @classmethod
    def from_json(cls, data: DeploymentRepresentation) -> Deployment:
        """
        Create Deployment from JSON representation

        :param data: JSON representation of Deployment
        :return: Deserialized Deployment
        """
        instance = cls.__new__(cls)

        instance.node = Node.from_json(data["node"])
        instance.prepared = data["prepared"]
        instance.executor_id = data["executor_id"]
        instance.error = Exception(data["error"]) if data["error"] else None
        instance.pipeline = b64decode(data["pipeline"])
        instance.keep_alive_timeout_minutes = data["keep_alive_timeout_minutes"]
        instance.cleanup = data.get("cleanup", True)
        instance.environment_definition = getattr(
            netunicorn.base.environment_definitions, data["environment_definition_type"]
        ).from_json(data["environment_definition"])
        return instance



"""
Experiment-related entities and classes.
"""
from __future__ import annotations

import base64
import copy
from dataclasses import dataclass
from enum import Enum
from typing import Dict, Iterator, List, Optional, Sequence, Tuple, Union, cast

from returns.pipeline import is_successful
from returns.result import Result

from .deployment import Deployment
from .nodes import Node, Nodes
from .pipeline import Pipeline
from .types import (
    DeploymentExecutionResultRepresentation,
    ExperimentExecutionInformationRepresentation,
    ExperimentRepresentation,
    PipelineResult,
)
from .utils import LogType


class ExperimentStatus(Enum):
    """
    Represents a status of an experiment.
    """

    UNKNOWN = 0
    """
    Unknown status.
    """

    PREPARING = 1
    """
    Experiment is under preparation.
    """

    READY = 2
    """
    Experiment is prepared and ready to be started.
    """

    RUNNING = 3
    """
    Experiment is running.
    """

    FINISHED = 4
    """
    Experiment is finished.
    """

    def __json__(self) -> int:
        return self.value

    @classmethod
    def from_json(cls, value: int) -> ExperimentStatus:
        """
        Converts a JSON representation of an experiment status to an instance of ExperimentStatus.

        :param value: experiment status value.
        :return: an instance of ExperimentStatus.
        """
        return cls(value)


class Experiment:
    """
    Represents an experiment that contains a mapping of pipelines to nodes.

    :param deployment_context: deployment context to be used by connectors.
    """

    def __init__(
        self, deployment_context: Optional[Dict[str, Dict[str, str]]] = None
    ) -> None:
        self.deployment_map: List[Deployment] = []
        """
        a list of deployments
        """

        self.deployment_context: Optional[
            Dict[str, Dict[str, str]]
        ] = deployment_context
        """
        A dictionary that contains a context for deployments.
            Context is to be provided by connectors. 
            Format: {connector_name: {key: value}}
        """

    def append(self, node: Node, pipeline: Pipeline) -> Experiment:
        """
        Append a new deployment (mapping of pipeline to a node) to the experiment.

        :param node: a node to deploy the pipeline to.
        :param pipeline: a pipeline to deploy.
        :return: self.
        """

        self.deployment_map.append(Deployment(node, pipeline))
        return self

    def map(self, pipeline: Pipeline, nodes: Sequence[Node]) -> Experiment:
        """
        Map a pipeline to a sequence of nodes.

        :param pipeline: a pipeline to deploy.
        :param nodes: a sequence of nodes to deploy the pipeline to.
        :return: self.
        """

        if isinstance(nodes, Nodes):
            raise TypeError("Expected sequence of nodes, got Nodes instead")

        for node in nodes:
            if not isinstance(node, Node):
                raise TypeError(f"Expected sequence of nodes, got {type(node)} instead")
            self.append(node, pipeline)
        return self

    def __json__(self) -> ExperimentRepresentation:
        return {
            "deployment_map": [x.__json__() for x in self.deployment_map],
            "deployment_context": self.deployment_context,
        }

    @classmethod
    def from_json(cls, data: ExperimentRepresentation) -> Experiment:
        """
        Creates an instance of Experiment from a JSON representation.

        :param data: a JSON representation of an experiment.
        :return: an instance of Experiment.
        """
        instance = cls.__new__(cls)
        instance.deployment_map = [
            Deployment.from_json(x) for x in data["deployment_map"]
        ]
        instance.deployment_context = data.get("deployment_context")
        return instance

    def __getitem__(self, item: int) -> Deployment:
        """
        Returns a deployment by index.

        :param item: an index of a deployment.
        :return: a deployment.
        """
        return self.deployment_map[item]

    def __iter__(self) -> Iterator[Deployment]:
        """
        Returns an iterator over deployments.

        :return: an iterator over deployments.
        """
        return iter(self.deployment_map)

    def __len__(self) -> int:
        """
        Returns a number of deployments in the experiment.

        :return: a number of deployments in the experiment.
        :meta public:
        """
        return len(self.deployment_map)

    def __str__(self) -> str:
        return "\n".join([f" - {x}" for x in self.deployment_map])

    def __repr__(self) -> str:
        return self.__str__()

    def __add__(self, other: Experiment) -> Experiment:
        """
        Concatenates two experiments resulting in a union of deployments.

        :param other: an experiment to concatenate with.
        :return: a new experiment.
        """

        new_map = copy.deepcopy(self)
        new_map.deployment_map.extend(other.deployment_map)
        if other.deployment_context:
            if new_map.deployment_context is None:
                new_map.deployment_context = {}
            new_map.deployment_context.update(other.deployment_context)
        return new_map


class DeploymentExecutionResult:
    """
    Stores a result (or ongoing information) of a deployment execution.

    :param node: a node that was used for deployment.
    :param serialized_pipeline: a serialized pipeline.
    :param result: a result of a deployment execution.
    :param error: an error message if deployment failed.
    """

    def __init__(
        self,
        node: Node,
        serialized_pipeline: bytes,
        result: Optional[bytes],
        error: Optional[str] = None,
    ):
        self.node: Node = node
        """
        a node that was used for deployment
        """

        self._pipeline: bytes = serialized_pipeline
        """
        a serialized pipeline
        """

        self._result: Optional[bytes] = result
        """
        Deployment execution result
        """

        self.error: Optional[str] = error
        """
        An error message if deployment failed.
        """

    @property
    def pipeline(self) -> Pipeline:
        """
        Returns a pipeline that was used for deployment.

        :return: a pipeline that was used for deployment.
        """
        import cloudpickle

        return cast(Pipeline, cloudpickle.loads(self._pipeline))

    @property
    def result(
        self,
    ) -> Optional[Tuple[Result[PipelineResult, PipelineResult], LogType]]:
        """
        Returns a result of a deployment execution and logs.

        :return: a tuple of (execution result, logs).
        """
        import cloudpickle

        return cloudpickle.loads(self._result) if self._result else None

    def __str__(self) -> str:
        text = "DeploymentExecutionResult:\n  Node: {self.node}\n"
        result = self.result
        if result:
            text += f"  Result: {type(result[0])}\n"
            if not is_successful(result[0]):
                text += f"   {result[0]}\n"
            else:
                for task_id, task_result in result[0].unwrap().items():
                    text += f"    {task_id}: {task_result}\n"
            text += f"  Logs:\n"
            for line in result[1]:
                text += f"    {line}"
        if self.error:
            text += f"  Error: {self.error}\n"
        text += "\n"
        return text

    def __repr__(self) -> str:
        return self.__str__()

    def __json__(self) -> DeploymentExecutionResultRepresentation:
        return {
            "node": self.node.__json__(),
            "pipeline": base64.b64encode(self._pipeline).decode("utf-8"),
            "result": base64.b64encode(self._result).decode("utf-8")
            if self._result
            else None,
            "error": self.error,
        }

    @classmethod
    def from_json(
        cls, data: DeploymentExecutionResultRepresentation
    ) -> DeploymentExecutionResult:
        """
        Returns an instance of DeploymentExecutionResult from a JSON representation.

        :param data: a JSON representation of a deployment execution result.
        :return: an instance of DeploymentExecutionResult.
        """
        return cls(
            Node.from_json(data["node"]),
            base64.b64decode(data["pipeline"]),
            base64.b64decode(data["result"]) if data["result"] else None,
            data["error"],
        )


@dataclass(frozen=True)
class ExperimentExecutionInformation:
    """
    Stores information about an experiment execution.

    :param status: a status of an experiment execution.
    :param experiment: a definition of an experiment.
    :param execution_result: a result of an experiment execution.
    """

    status: ExperimentStatus
    """
    Describes the status of an experiment execution.
    """

    experiment: Optional[Experiment]
    """
    Stores the definition of the experiment.
    """

    execution_result: Union[None, Exception, List[DeploymentExecutionResult]]
    """
    Stores either an Exception if experiment execution failed or a list of deployment execution results.
    """

    def __str__(self) -> str:
        text = (
            f"ExperimentExecutionInformation:\n"
            f"status: {self.status}\n"
            f"experiment: \n"
            f"{self.experiment}\n"
            f"execution_result:\n"
            f"{self.execution_result}\n"
        )
        return text

    def __repr__(self) -> str:
        return self.__str__()

    def __json__(self) -> ExperimentExecutionInformationRepresentation:
        execution_result: Union[
            None, str, List[DeploymentExecutionResultRepresentation]
        ] = None
        if isinstance(self.execution_result, list):
            execution_result = [x.__json__() for x in self.execution_result]
        elif isinstance(self.execution_result, Exception):
            execution_result = str(self.execution_result.__reduce__())
        return {
            "status": self.status.__json__(),
            "experiment": self.experiment.__json__() if self.experiment else None,
            "execution_result": execution_result,
        }

    @classmethod
    def from_json(
        cls, data: ExperimentExecutionInformationRepresentation
    ) -> ExperimentExecutionInformation:
        """
        Returns an instance of ExperimentExecutionInformation from a JSON representation.

        :param data: a JSON representation of an experiment execution information.
        :return: an instance of ExperimentExecutionInformation.
        """

        status = ExperimentStatus.from_json(data["status"])
        experiment = (
            Experiment.from_json(data["experiment"]) if data["experiment"] else None
        )
        execution_result_data = data["execution_result"]
        execution_result: Union[None, Exception, List[DeploymentExecutionResult]] = None
        if execution_result_data:
            if isinstance(execution_result_data, list):
                execution_result = [
                    DeploymentExecutionResult.from_json(x)
                    for x in execution_result_data
                ]
            else:
                execution_result = Exception(execution_result_data)
        return cls(status, experiment, execution_result)



"""
Abstrcations for Pipeline representation.
"""

from __future__ import annotations

import uuid
from copy import deepcopy
from typing import Collection, List, Optional, Set, Union
from warnings import warn

from .environment_definitions import DockerImage, EnvironmentDefinition
from .task import TaskElement

PipelineElement = Union[TaskElement, Collection[TaskElement]]


class Pipeline:
    """
    Pipeline is a class that takes a tuple of Tasks and executes them in order.
    Each element in the tuple should be either a Task or a tuple of Tasks.
    Pipeline will execute elements in order and return the combined result of all tasks.
    If element is a tuple of tasks, these tasks would be executed in parallel.

    | The result of pipeline execution would be one of the following:
    | - Success: if all tasks succeed, then Success is returned
    | - Failure: if any task fails, then Failure is returned

    | Returning object will always contain a Result object for each task executed.
    | If early_stopping is set to True, then any task after first failed wouldn't be executed.

    :param tasks: tasks (ordered by stages) to be executed
    :param early_stopping: whether to stop executing tasks after first failure
    :param report_results: whether executor should connect core services to report pipeline results in the end
    :param environment_definition: environment definition for the pipeline
    """

    def __init__(
        self,
        tasks: Collection[PipelineElement] = (),
        early_stopping: bool = True,
        report_results: bool = True,
        environment_definition: Optional[EnvironmentDefinition] = None,
    ):
        self.name: str = str(uuid.uuid4())
        """
        Pipeline name.
        """

        self.task_names: Set[str] = set()
        """
        Task names in the pipeline, used for dictionary of results.
        """

        self.early_stopping: bool = early_stopping
        """
        Whether to stop executing tasks after first failure.
        """

        self.tasks: List[List[TaskElement]] = []
        """
        Tasks stages to be executed.
        """

        self.report_results: bool = report_results
        """
        Whether executor should connect core services to report pipeline results in the end.
        """

        self.environment_definition: EnvironmentDefinition = (
            environment_definition or DockerImage()
        )
        """
        Environment definition for the pipeline.
        """

        for element in tasks:
            self.then(element)

    @staticmethod
    def __element_to_stage(element: PipelineElement) -> List[TaskElement]:
        if not isinstance(element, Collection):
            element = [element]
        if not isinstance(element, list):
            element = list(element)
        return element

    def __check_tasks_names(self, stage: List[TaskElement]) -> None:
        for task in stage:
            if task.name in self.task_names:
                warn(
                    f"Task with name {task.name} already exists in the current pipeline {self.__str__()}. "
                    "Please, note that execution results of these tasks could be mixed or overwritten."
                )
            self.task_names.add(task.name)

    def then(self, element: PipelineElement) -> Pipeline:
        """
        Add a task or list of tasks as a separate stage to the end of the pipeline.

        :param element: a task or tuple of tasks to be added
        :return: self
        """
        element = self.__element_to_stage(element)
        self.__check_tasks_names(element)
        self.tasks.append(element)
        return self

    def copy(self) -> Pipeline:
        """
        Return a copy of the pipeline.

        :return: a copy of the pipeline
        """
        return Pipeline(deepcopy(self.tasks), self.early_stopping)

    def __str__(self) -> str:
        return f"Pipeline({self.name}): {self.tasks}"

    def __repr__(self) -> str:
        return str(self)
